' File: FindMaterialsElements_wNoCTE.BAS
' Version 03
' Date: May 19, 2017
' Author: Ryan Blelloch, ATA Engineering
' Support: kurt.knutson@ata-e.com
' www.ata-plmsoftware.com
'
' Description:  This API finds all materials and elements that might not have a CTE or reference
'		temperature and stores them in groups for the user to examine. Message window messages
' 		identify bad material and element totals, and the data table shows materials
'		missing CTEs and/or reference temperatures.  Materials are flagged if they have a CTE 
'		value or reference temperature of 0.  Rigid elements are flagged if their rigid 
'		thermal expansion is 0.  A maximum of five groups may be created: 
'		1) Material with 0 CTE
'		2) Material with 0 reference temperature
'		3) Material with 0 CTE and 0 reference temperature
'		4) RBE2s with 0 rigid thermal expansion
'		5) RBE3s with 0 rigid thermal expansion
'
'
' Version01	This FEMAP macro finds all materials and elements that might not have a CTE or ref. temperature,
' 		and stores them groups for the engineer to check on. If a material has a 0 in one of its CTE values,
' 		or a 0 as its reference temperature, it is placed in the group. Rigid elements are checked if their
' 		rigid thermal expansion is 0, and are placed in another group if so.
'
' Version02     Improved version seperates materials w/ no CTE and materials with no ref. temperature
'		Also a third group that has both. It also seperates RBE2 from RBE3 and only searches through
'		rigid elements, rather than all elements in the model.
'		The maximum amount of groups created is 5. The minimum is 0.
'
' Version03     Removed printed messages to message window for individual elements
'		Changed message box messages for total bad matls and total bad elems to message window messages
'		Added data table print for all materials with missing CTE or Reference Temp
'		data table shows true/false for each material missing CTE or Reference Temp
'		only materials with at least one missing are printed to the data table
'
'
'Copyright (c) <2017> <ATA Engineering, Inc.>
'
'Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
'associated documentation files (the "Software") to use, copy, modify, merge, and/or distribute
'royalty-free copies of the Software, and to permit persons to whom the Software is furnished
'to do so, subject to the following conditions:
'The above copyright notice and this permission notice shall be included in all copies or
'substantial portions of the Software.
'THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
'FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
'WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
'OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'

Sub Main
    Dim App As femap.model
    Set App = feFemap()

    Dim allMat As femap.Set
    Set allMat = App.feSet
    allMat.Clear

    'Used to load in the current material for testing
    Dim cur_Mat As femap.Matl
    Set cur_Mat = App.feMatl

    'used to load in the current element for testing
    Dim cur_Elem As femap.Elem
    Set cur_Elem = App.feElem


    'used to loop through all materials and elements
    Dim loopSet As femap.Set
    Set loopSet = App.feSet


    ''''''''''''ALL COUNTERS FOR MATERIALS AND ELEMENTS'''''''''''
    'Used to count the number of bad materials
    Dim badMat_Counter As Integer
    badMat_Counter = 0

    'Used to count the number of bad elements
    Dim badElem2_Counter As Integer
    badElem2_Counter = 0

    'Used to count the number of bad elements
    Dim badElem3_Counter As Integer
    badElem3_Counter = 0
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

    '''''''''''''''''ALL SETS TO CREATE GROUPS'''''''''''''''''''''
    'Used to store all the materials with possibly no CTE
    Dim matNoCTE_Set As femap.Set
    Set matNoCTE_Set = App.feSet
    matNoCTE_Set.Clear

    'Used to store all the materials with no reference temperature
    Dim matNoRefTemp_Set As femap.Set
    Set matNoRefTemp_Set = App.feSet
    matNoRefTemp_Set.Clear

    'Used to store all materials with no CTE and no reference temperature
    Dim matNoCTEorRefTemp_Set As femap.Set
    Set matNoCTEorRefTemp_Set = App.feSet
    matNoCTEorRefTemp_Set.Clear

    'Used to store all the rigid elements with no Rigid Thermal Expansion
    Dim elem2NoThermExpd_Set As femap.Set
    Set elem2NoThermExpd_Set = App.feSet
    elem2NoThermExpd_Set.Clear

     'Used to store all the rigid elements with no Rigid Thermal Expansion
     Dim elem3NoThermExpd_Set As femap.Set
     Set elem3NoThermExpd_Set = App.feSet
     elem3NoThermExpd_Set.Clear
     '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

      '''''''''''''''''DATA TABLE TO PERSENT BAD MATERIALS'''''''''''''''''''''
      Dim matlDataTablePrint As Boolean
      matlDataTablePrint = True

      Dim matlDataTable As femap.DataTable
      Set matlDataTable = App.feDataTable
      matlDataTable.Visible = False

      'combination set for all bad materials
      Dim allBadMatl As femap.Set
      Set allBadMatl = App.feSet
      allBadMatl.Clear

      'ID values for materials
      Dim matlIDs() As Long

      'string values for material names column
      Dim matlNames() As String
      Dim matlNames_colIndex As Long

      'boolean values for missing CTE column
      Dim zeroCTE() As Boolean
      Dim zeroCTE_colIndex As Long

	'boolean values for missing Reference Temp column
      Dim zeroRtemp() As Boolean
      Dim zeroRtemp_colIndex As Long
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

    'add all materials to testing set for testing
    loopSet.AddAll(FT_MATL)

    While loopSet.Next
    	cur_Mat.Get( loopSet.CurrentID )

    	'Check to see if any CTE coefficients are greater than 0 when they should be activated
    	If( checkFor0CTE_Mat( cur_Mat, cur_Mat.type) = False ) Then
    	Else 'has no CTE
    		matNoCTE_Set.Add(loopSet.CurrentID)
    		badMat_Counter = badMat_Counter + 1
    		'App.feAppMessage(FCM_NORMAL, "Material: " & CStr(cur_Mat.ID) & ". " & cur_Mat.title & " has a 0 in its CTE")
    	End If

    	'Check to see if reference temperature is greater than 0
    	If( ( cur_Mat.RefTemp > 0 ) ) Then
    	Else 'has no reference material
    		matNoRefTemp_Set.Add(loopSet.CurrentID)
    		badMat_Counter = badMat_Counter + 1
    		'App.feAppMessage(FCM_NORMAL, "Material: " & CStr(cur_Mat.ID) & ". " &  cur_Mat.title & " has ref temp = 0")
    	End If

    Wend


    'generate data table with all the bad materials
    If((matlDataTablePrint) And ((matNoCTE_Set.Count + matNoRefTemp_Set.Count) > 0)) Then

    	'Make the data table visible if it is not already
	    If Not(matlDataTable.Visible) Then
	    	App.feAppManagePanes("Data Table", 1)
	    End If

	    'Unlocked the data table if it is currently unlocked
	    If(matlDataTable.Locked) Then
	    	matlDataTable.Lock(False)
	    End If

		allBadMatl.AddSet(matNoCTE_Set.ID)
	    allBadMatl.AddSet(matNoRefTemp_Set.ID)
	    fillArrayWithIDsNamesOfSet( cur_Mat, allBadMatl, matlIDs, matlNames )

	    'fill missingCTE array
	    fillArrayUsingSubSet(zeroCTE, allBadMatl, matNoCTE_Set)
	    'fill missingRtemp array
	    fillArrayUsingSubSet(zeroRtemp, allBadMatl, matNoRefTemp_Set)
	    'Add Column for material names
	    matlDataTable.AddColumn(True, False, FT_MATL, 0, "Matl Name", FCT_STRING, allBadMatl.Count, matlIDs, matlNames,matlNames_colIndex)
	    matlDataTable.SetColumnWidth( matlNames_colIndex, 130 )
	    'Add Column for missing CTE
	    matlDataTable.AddColumn(False, False, FT_MATL, 0, "0 in CTE", FCT_BOOL, allBadMatl.Count, matlIDs, zeroCTE,zeroCTE_colIndex)
	    matlDataTable.SetColumnWidth( zeroCTE_colIndex, 50 )
	    'Add Column for missing Reference Temp
	    matlDataTable.AddColumn(False, False, FT_MATL, 0, "Ref Temp = 0", FCT_BOOL, allBadMatl.Count, matlIDs, zeroRtemp,zeroRtemp_colIndex)
		matlDataTable.SetColumnWidth( zeroRtemp_colIndex, 78 )
		matlDataTable.Visible = True
    End If


    'Find which materials were missing both
    matNoCTEorRefTemp_Set.AddCommon( matNoCTE_Set.ID , matNoRefTemp_Set.ID )
    badMat_Counter = badMat_Counter - matNoCTEorRefTemp_Set.Count
    'Remove all the common materials from the individual first to sets
    matNoCTE_Set.RemoveSet( matNoCTEorRefTemp_Set.ID )
    matNoRefTemp_Set.RemoveSet( matNoCTEorRefTemp_Set.ID )

    'Message to user on how many materials were found
    'mG = App.feAppMessageBox( 0, "Macro found" & Str$(badMat_Counter) & " bad materials")
    mG = App.feAppMessage(FCM_NORMAL, "Macro found" & Str$(badMat_Counter) & " bad materials")

    'Generate a group of all materials with no CTE
    If(matNoCTE_Set.Count) Then
        buildGroup_WithSet(App, matNoCTE_Set.ID, femap.zDataType.FT_MATL, "Materials: No CTE")
    End If
    'Generate a group of all materials with no Reference Temp
    If (matNoRefTemp_Set.Count) Then
        buildGroup_WithSet(App, matNoRefTemp_Set.ID, femap.zDataType.FT_MATL, "Materials: No Reference Temp")
    End If
    'Generate a group of all materials with neither
    If (matNoCTEorRefTemp_Set.Count > 0) Then
        buildGroup_WithSet(App, matNoCTEorRefTemp_Set.ID, femap.zDataType.FT_MATL, "Materials: No CTE or Reference Temp")
    End If

    're-use loop set for rigid element test
    loopSet.Clear
    loopSet.AddRule( 29, FGD_ELEM_BYTYPE ) 'look at only the rigid elements

    'loop through all of the elements
    While loopSet.Next

    	'get the current element
    	cur_Elem.Get( loopSet.CurrentID )

    	'check to see if the elem is of rigid type
    	If(cur_Elem.type = FET_L_RIGID) Then

    		'check to see if the element has a rigid thermal expansion
	    	If(cur_Elem.RigidThermalExpansion >  0.0) Then
	    	Else 'Has no CTE

	    		'Decide if it is an RBE2 or and RBE3
    			If  ( cur_Elem.RigidInterpolate = False ) Then 'must be RBE2
    				elem2NoThermExpd_Set.Add( loopSet.CurrentID )
    				badElem2_Counter = badElem2_Counter + 1
    			Else 'must be RBE3
    				elem3NoThermExpd_Set.Add( loopSet.CurrentID )
    				badElem3_Counter = badElem3_Counter + 1
    			End If
	    	End If
    	End If
    Wend

    'Tell user how many of each bad material there was
    'mG = App.feAppMessageBox( 0, "Macro found" & Str$(badElem2_Counter) & " bad RBE2 and" & Str$(badElem3_Counter) & " bad RBE3")
    mG = App.feAppMessage(FCM_NORMAL, "Macro found" & Str$(badElem2_Counter) & " bad RBE2 and" & Str$(badElem3_Counter) & " bad RBE3")

    'create group with the remaining elements
    If ( elem2NoThermExpd_Set.Count > 0 ) Then
        buildGroup_WithSet(App, elem2NoThermExpd_Set.ID, femap.zDataType.FT_ELEM, "RBE2: No CTE")
    End If
    If ( elem2NoThermExpd_Set.Count > 0 ) Then
        buildGroup_WithSet(App, elem3NoThermExpd_Set.ID, femap.zDataType.FT_ELEM, "RBE3: No CTE")
    End If

End Sub

'Checks the coefficients of thermal expansion for 0's where they should not be, based on material type passed in, returns true if found
Public Function checkFor0CTE_Mat( ByRef cM As femap.Matl, ByVal matType As Integer) As Boolean

	Dim foundZero As Boolean
	foundZero = False

	Select Case (matType)

	Case 0, 1, 2 'Isotropic, Orthotropic (2D), Orthotropic (3D)
		If( (cM.Alphaxx = 0.)  Or (cM.Alphayy = 0.) Or (cM.Alphazz = 0.) ) Then
		    foundZero = True
	    End If
	Case 3, 4 'Anisotropic (2D), Anisotropic (3D)
		If( (cM.Alphaxx = 0.) Or (cM.Alphaxy = 0.) Or (cM.Alphayy = 0.) Or (cM.Alphayz = 0.) Or (cM.Alphazx = 0.) Or (cM.Alphazz = 0.) ) Then
		    foundZero = True
	    End If
	Case 5, 7 'Hyperelastic, Fluid
		If( (cM.Alphaxx = 0) ) Then
		    foundZero = True
	    End If
	Case 6 'General
		'Should never happen
	End Select

	'return if zero coefficient was found where it shouldn't be
	checkFor0CTE_Mat = foundZero

End Function

'Builds a new group of elements
Public Sub buildGroup_WithSet( ByRef App As femap.model, ByVal setID As Integer, ByVal setType As femap.zDataType, ByVal group_title As String)

	Dim cG As femap.Group
	Set cG = App.feGroup

	Dim gID As Long

	cG.Last
    gID = cG.ID + 1
    cG.Get( gID )

    'Since new group is a copy of the old one, we want to remove previous entities that may have been in it
    cG.RangeDeleteAll(-1)

    cG.title = group_title
    cG.SetAdd(setType, setID)

    cG.Put( gID )

End Sub

'Builds boolean array of Set based on finding it in subset
Public Sub fillArrayUsingSubSet( ByRef arr() As Boolean, ByRef mainSet As femap.Set, ByRef subSet As femap.Set)

	'Redimension array to fit size of set
	ReDim arr(mainSet.Count - 1)
	Dim arrCnt As Long
	arrCnt = 0

	'Loop through main set in order, checking whether or not ID is found in subset
	mainSet.Reset
	While mainSet.Next
		arr(arrCnt) = subSet.IsAdded(mainSet.CurrentID)
		arrCnt = arrCnt + 1
	Wend

End Sub

'Fills two arrays with values of set
'dummyMatl is femap.Matl needed to access title
'tSet is femap.Set with material information
'arrID is Long array to store all IDs stored in set in order
'arrNames is String array to store all titles stored in set in order
Public Sub fillArrayWithIDsNamesOfSet(ByRef dummyMatl As femap.Matl, ByRef tSet As femap.Set, ByRef arrID() As Long, ByRef arrName() As String )

	ReDim arrID(tSet.Count - 1)
	ReDim arrName(tSet.Count - 1)
	Dim arrCnt As Long
	arrCnt = 0
	tSet.Reset
	While tSet.Next
		dummyMatl.Get(tSet.CurrentID)
		arrID(arrCnt) = dummyMatl.ID
		arrName(arrCnt) =dummyMatl.title
		arrCnt = arrCnt + 1
	Wend

End Sub
